// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//						(c) 1997 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	Lightcurve.cc --- lightcurve class
//
//	DESCRIPTION:
//		Lightcurve
//
//	CHANGE HISTORY:
//		AI, 2007-05-30: 1.0.0: - first version
//
//	AUTHOR:
//		A. Ibarra, Wed May 30 16:32:18 GMT 2007
//
//	REFERENCES:
//
//
//

#include "Lightcurve.h"
#include "param.h"
#include "DataFilter.h"
#include <valarray>
#include <iomanip>
#include <cmath>
#include <Random.h>
#include <fstream>
/* AI gcc4 porting
 *‘INT_MIN’ was not declared in this scope
 * => #include <climits>
 */
#include <climits>

using namespace std;

inline int tint( double x )
{
  return x > INT_MIN+1 ? x < INT_MAX-1 ? (int)x : INT_MAX-1 : INT_MIN+1;
}

Lightcurve::Lightcurve(double binSize) :
  _binSize(binSize)
{
  _startTime = -1.0;
  _endTime = -1.0;
  _ccdNr=0;
  _numberLcAdded = 1.0;
  _totalCounts = 0.0;
  _backScale = 1.0;
  _backScaleWeighted = 0.0;
}

Lightcurve::Lightcurve(double binSize, int ccdNr) :
  _binSize(binSize)
{
  _startTime = -1.0;
  _endTime = -1.0;
  _ccdNr = ccdNr;
  _numberLcAdded = 1.0;
  _totalCounts = 0.0;
  _backScale = 1.0;
  _backScaleWeighted = 0.0;
}
void Lightcurve::setTimeLimits(const Column * col)
{

  vector<double> timeLimits = _lcUtils.getRange(col);

  _startTime = timeLimits[0];
  _endTime = timeLimits[1];
}


void Lightcurve::setTimeLimits(const Column * startCol,const Column * stopCol)
{
  vector<double> startTimeLimits = _lcUtils.getRange(startCol);
  vector<double> stopTimeLimits = _lcUtils.getRange(stopCol);

  /* Reset the initial values */
  if (_startTime == -1.0)
    _startTime = startTimeLimits[0];
  if (_endTime == -1.0)
    _endTime =  stopTimeLimits[1];

  if (_startTime > startTimeLimits[0])
    _startTime = startTimeLimits[0];
  if (_endTime < stopTimeLimits[1])
    _endTime = stopTimeLimits[1];


}

void Lightcurve::setTimeLimits(double startTime, double endTime)
{
  _startTime = startTime;
  _endTime = endTime;

}

vector<double> Lightcurve::getTimeLimits()
{
  vector<double> timeLimits;
  timeLimits.push_back(_startTime);
  timeLimits.push_back(_endTime);
  return timeLimits;
}


void Lightcurve::createTimeGrid()
{

  _lcUtils.createGrid(_binSize,_startTime,_endTime);
  _timeGrid.resize(_lcUtils.getNBins()+1,0);
  _timeGrid = _lcUtils.getTimeGrid();
}

int Lightcurve::getCcdNr()
{
  return _ccdNr;
}

double  Lightcurve::getTotalCounts()
{
  return _totalCounts;
}


void Lightcurve::setExposuInfo(const Table *tab, const Table *gtiTab, const Column *col)
{

  /*Check if the number of rows in the STDGTI extension and EXPOSU extension are the same
    if not, we stop the task and we warn the users about the inconsistency of the ODF.
    The cause of this problem is problaby a JUMP in the frame counter of the AUX and SPE 
    files */
  if ( tab->rows() != gtiTab->rows())
    errstr << "This ODF constains difference number of rows in the GTI and EXPOSU extensions. " 
	   << " This should not happen. Please contact XMM-Newton SOC" 
	   << Sas::error("GTI&EXPOSU Inconsistency");

  /* Exposu Time and integration time info */
  const Column *expoTimeCol = tab->column("TIME",Dal::Read);
  ColumnData *cExpoTimeData = expoTimeCol->data();
  real64 *expoTimeData = cExpoTimeData->real64Data();

  const Column *expoTimeDelCol = tab->column("TIMEDEL",Dal::Read);
  ColumnData *cExpoTimeDelData = expoTimeDelCol->data();
  real32 *expoTimeDelData = cExpoTimeDelData->real32Data();
  /* Exposu Time and integration time info */

  /* GTI Table */
  const Column *gtStartCol = gtiTab->column("START",Dal::Read);
  ColumnData *cGtiStartData = gtStartCol->data();
  real64 *gtiStartData = cGtiStartData->real64Data();

  const Column *gtStopCol = gtiTab->column("STOP",Dal::Read);
  ColumnData *cGtiStopData = gtStopCol->data();
  real64 *gtiStopData = cGtiStopData->real64Data();
  /* GTI Table */

  ColumnData *cData = col->data();
  real64 *tData = cData->real64Data();

  /*
   * Create the map with the exposu time and the
   * intregration time
   */
  for (unsigned int i=0 ; i < expoTimeCol->rows() ; i++)
    _expoTimeDelMap[expoTimeData[i]] = expoTimeDelData[i];
  /*
   * Create the map with the time exposu values
   */
  for (unsigned int i=0 ; i < expoTimeCol->rows() ; i++)
    _expoCountMap[expoTimeData[i]] = 0;
  /*
   * How many events do we have for each frame time?
   * Add this events to the _expoCountMap
   */
  map<double,int>::iterator it;

  for (unsigned int i=0 ; i < col->rows() ; i++)
    {
	  it = _expoCountMap.find(tData[i]);
      if (it!=_expoCountMap.end())
    	  (*it).second = (*it).second + 1;

    }

  /*
   * Create the exposure map together with the gti values
   */
  vector<double> gti;
  for (unsigned int i=0 ; i < expoTimeCol->rows() ; i++)
    {
      gti.push_back(gtiStartData[i]);
      gti.push_back(gtiStopData[i]);
      _expoGtiMap.insert(make_pair(expoTimeData[i], gti));
      //_expoGtiMap[expoTimeData[i]] = gti;
      gti.clear();
    }


  delete cData;
  delete cGtiStartData;
  delete cGtiStopData;
  delete cExpoTimeDelData;
  delete cExpoTimeData;
}

void Lightcurve::createLc()
{
  _counts.resize(_timeGrid.size()-1,0);
  _errCounts.resize(_timeGrid.size()-1,0);
  _fracExp.resize(_timeGrid.size()-1,0);

  //Create also the countRate vector
  _countRate.resize(_timeGrid.size()-1,0);
  _errCountRate.resize(_timeGrid.size()-1,0);

  _realCounts.resize(_timeGrid.size()-1,0);

  std::map<double,vector<double> >::iterator expoGtiMapIt = _expoGtiMap.begin();

  //First bin
  while (( ((*expoGtiMapIt).second)[0] < _timeGrid[0]) &&
	 ( ((*expoGtiMapIt).second)[1] <= _timeGrid[0]) )
    ++expoGtiMapIt;

  // First bin!!
  //  Fraction of counts that go to the current bin
  if ( ( expoGtiMapIt != _expoGtiMap.end()) &&
		  ((((*expoGtiMapIt).second)[0] < _timeGrid[0] ) &&
    	  (((*expoGtiMapIt).second)[1] > _timeGrid[0] ) ) )
    {

      double firstFracTime = (_timeGrid[0] - ((*expoGtiMapIt).second)[0]) /
						  (((*expoGtiMapIt).second)[1]- ((*expoGtiMapIt).second)[0]);

      for (int ev = 0 ; ev < _expoCountMap[(*expoGtiMapIt).first] ; ev++)
      {
    	  if(randGen.rand() >= firstFracTime)
    		  _counts[0]++;
      }

      _fracExp[0] = _fracExp[0] +
		  (1.0 - firstFracTime) * _expoTimeDelMap[(*expoGtiMapIt).first];

      if ( _expoGtiMap.end() != expoGtiMapIt )
    	  ++expoGtiMapIt;
    }



  for (unsigned int i = 0 ; i < _timeGrid.size()-2 ; i++)
    {

      while ( ( expoGtiMapIt != _expoGtiMap.end()) && (
    		  ( ((*expoGtiMapIt).second)[0] == _timeGrid[i])  ||
    		  ( ((*expoGtiMapIt).second)[1] == _timeGrid[i+1]) ||
    		  ( ( ((*expoGtiMapIt).second)[0] >= _timeGrid[i])  &&
    				  (((*expoGtiMapIt).second)[1] <= _timeGrid[i+1])  ) ) )
	{

	  // Fraction of counts that go to the current bin
	  _counts[i] = _counts[i]+ _expoCountMap[(*expoGtiMapIt).first];
	  _fracExp[i] = _fracExp[i] + _expoTimeDelMap[(*expoGtiMapIt).first];

	  if ( expoGtiMapIt != _expoGtiMap.end())
	    ++expoGtiMapIt;
	}

      //Check if we have to add something to the current bin
      if (  ( expoGtiMapIt != _expoGtiMap.end()) &&
    		  ( ((*expoGtiMapIt).second)[0] < _timeGrid[i+1] ) )
      {

    	  double fracTime = (_timeGrid[i+1] - ((*expoGtiMapIt).second)[0]) /
    			  (((*expoGtiMapIt).second)[1]- ((*expoGtiMapIt).second)[0]);

    	  for (int ev = 0 ; ev < _expoCountMap[(*expoGtiMapIt).first] ; ev++)
    	  {
    		  if(randGen.rand() <= fracTime)
    			  _counts[i]++;
    		  else
    			  _counts[i+1]++;
    	  }

    	  _fracExp[i] = _fracExp[i] +
    			  fracTime * _expoTimeDelMap[(*expoGtiMapIt).first];
    	  _fracExp[i+1] = _fracExp[i+1] +
    			  (1.0 - fracTime) * _expoTimeDelMap[(*expoGtiMapIt).first];

    	  if ( _expoGtiMap.end() != expoGtiMapIt )
    		  ++expoGtiMapIt;
      }
    }

  // Check the lastest bin
  while (( expoGtiMapIt != _expoGtiMap.end()) &&
	 (((*expoGtiMapIt).second)[1] <= _timeGrid[_timeGrid.size()-1]) )
    {
      _counts[_counts.size()-1] = _counts[_counts.size()-1] +	 _expoCountMap[(*expoGtiMapIt).first];
      _fracExp[_fracExp.size()-1] = _fracExp[_fracExp.size()-1] + _expoTimeDelMap[(*expoGtiMapIt).first];


      if ( _expoGtiMap.end() != expoGtiMapIt )
	++expoGtiMapIt;

    }


  errstr << "Calculating Count Rates... " << Sas::message(Msg::App,Msg::Sparse);


  for (unsigned int i = 0 ; i < _counts.size() ; i++)
	  _countRate[i] = (_counts[i]/_fracExp[i]);


  errstr << "done." <<Sas::message(Msg::App,Msg::Sparse);

}

void Lightcurve::showInfo()
{
  for (unsigned int i = 0 ; i < _counts.size() ; i++)
    {
      std::cout<< _timeGrid[i] + ((_timeGrid[i+1] - _timeGrid[i]) / 2.0) << " "
	       << _counts[i]/(_fracExp[i]/_numberLcAdded)  << " " << _fracExp[i]/_numberLcAdded << std::endl;
    }

}

void Lightcurve::showCCDInfo()
{
  for (unsigned int i = 0 ; i < _counts.size() ; i++)
    std::cout << _timeGrid[i] + ((_timeGrid[i+1] - _timeGrid[i]) / 2.0) << " "
	      << _counts[i]/_fracExp[i] << " " << _fracExp[i] << std::endl;
}


void Lightcurve::calculateWeightedBackScale(Lightcurve *bkgLc)
{

  /*
   * Total Source counts
   */
  double totalSrcCounts = 0.;
  double srcArea = 0.;
  double backScaleMean = 0.;
  //errstr << "CCD " << _ccdNr << ": Src counts: "<< _totalCounts <<" Src Area: "
//	 << _area << Sas::message(Msg::App,Msg::Sparse);
  for (unsigned int i = 0 ; i < _totalCCDCounts.size() ; i++)
    {
      errstr << "Next CCD: "<<   "Src counts "<< _totalCCDCounts[i] <<
	" Src Area " << _ccdArea[i] << Sas::message(Msg::App,Msg::Sparse);

      totalSrcCounts = totalSrcCounts + _totalCCDCounts[i];
      srcArea = srcArea + (_totalCCDCounts[i]*_ccdArea[i]);
      backScaleMean += _ccdArea[i]/bkgLc->_ccdArea[i];
    }
  srcArea = srcArea/totalSrcCounts;

  /*
   * Total background counts
   */
  double totalBkgCounts = 0.;
  double bkgArea = 0.;

  //errstr << "CCD " << bkgLc->_ccdNr << ": Bkg counts "<< bkgLc->_totalCounts <<" Bkg Area: "
//	 << bkgLc->getArea() << Sas::message(Msg::App,Msg::Sparse);
  for (unsigned int i = 0 ; i < bkgLc->_totalCCDCounts.size() ; i++)
    {
      errstr << "Next CCD Bkg counts "<< bkgLc->_totalCCDCounts[i] <<
	" Bkg Area " << bkgLc->_ccdArea[i] << Sas::message(Msg::App,Msg::Sparse);
      totalBkgCounts = totalBkgCounts + bkgLc->_totalCCDCounts[i];
      bkgArea = bkgArea + (bkgLc->_ccdArea[i]*bkgLc->_totalCCDCounts[i]);
    }
  bkgArea = bkgArea/totalBkgCounts;
  _backScaleWeighted = srcArea/bkgArea;

  errstr << "Mean BACKSCALE " << backScaleMean/_totalCCDCounts.size()
	 << Sas::message(Msg::App,Msg::Sparse);
}

void Lightcurve::add()
{

  for (unsigned int i = 0 ; i < _counts.size() ; i++)
    {
      _errCounts[i] = sqrt((_errCounts[i]*_errCounts[i] ));
      _errCountRate[i] = sqrt((_errCountRate[i]*_errCountRate[i]));
    }
  _totalCCDCounts.push_back(_totalCounts);
  _ccdArea.push_back(_area);

}


void Lightcurve::add(Lightcurve *lc)
{

  _numberLcAdded++;
  for (unsigned int i = 0 ; i < _counts.size() ; i++)
    {
      _countRate[i] = _countRate[i] + lc->_countRate[i];
      _counts[i] = _counts[i] + lc->_counts[i];
      _realCounts[i] = _realCounts[i] + lc->_realCounts[i];
      _fracExp[i] = _fracExp[i] + lc->_fracExp[i];
      _errCounts[i] = sqrt((_errCounts[i]*_errCounts[i] )+(lc->_errCounts[i]*lc->_errCounts[i]));
      _errCountRate[i] = sqrt((_errCountRate[i]*_errCountRate[i])+(lc->_errCountRate[i]*lc->_errCountRate[i]));
    }
  _totalCCDCounts.push_back(lc->_totalCounts);
  _ccdArea.push_back(lc->getArea());

}


void Lightcurve::calculateErrors()
{
  //Calculate the errors and the Exposure time
  for (unsigned int i = 0 ; i < _counts.size() ; i++)
    {
	  _errCounts[i] = sqrt(_realCounts[i]);
	  if (_realCounts[i] == 0)
		  _errCountRate[i] = 0.0;
	  else
		  _errCountRate[i] = sqrt(_realCounts[i])/(_realCounts[i] / _countRate[i]);
    }

}
void Lightcurve::substract(Lightcurve *bkgLC)
{

  for (unsigned int i = 0 ; i < _counts.size() ; i++)
    {
      _counts[i] = _realCounts[i] - bkgLC->_counts[i];
      //_errCounts[i] = sqrt(_realCounts[i] + _backScale*_backScale*bkgLC->_realCounts[i]);
      _countRate[i] = _countRate[i] - bkgLC->_countRate[i];
      _errCounts[i] = sqrt((_errCounts[i]*_errCounts[i]) + (bkgLC->_errCounts[i]*bkgLC->_errCounts[i]));
      _errCountRate[i] = sqrt((_errCountRate[i]*_errCountRate[i]) +
	  		      (bkgLC->_errCountRate[i]*bkgLC->_errCountRate[i]));
    }

}

void Lightcurve::setArea(double area)
{
  _area = area;
}

double Lightcurve::getArea()
{
  return _area;
}

void Lightcurve::accumCounts()
{
	 /*
	  * Number of gross source counts
	  */
	  for (unsigned int i = 0 ; i < _counts.size() ; i++) {
		_realCounts[i] = _counts[i];
	    _totalCounts = _totalCounts + _counts[i];
	  }
}

void Lightcurve::correctBackScale(Lightcurve *srcLc)
{
  srcLc->_backScale = srcLc->getArea()/_area;
  _backScale = srcLc->getArea()/_area;

  for (unsigned int i = 0 ; i < srcLc->_counts.size() ; i++) {
     _counts[i] = _realCounts[i] * (srcLc->getArea()/_area);
     _errCounts[i] = _errCounts[i]*(srcLc->getArea()/_area);
   }

  for (unsigned int i = 0 ; i < _countRate.size() ; i++) {
	  _errCountRate[i] = _errCountRate[i]*(srcLc->getArea()/_area);
	  _countRate[i] = _countRate[i] * (srcLc->getArea()/_area);
  }

}


void Lightcurve::setDefaultKeywords(std::vector<std::vector<std::string> > tmpFileNames, int srcNo)
{


  //Copy default attributes
  DataSet * srcSet = dataSetServer->open(tmpFileNames[0][0], Dal::Read);
  Table *tab = srcSet->table("EVENTS");
  _mission =tab->attribute("TELESCOP")->asString();

  if ( tmpFileNames[0][0].find("RGS1") != string::npos )
    _instrument = "RGS1";
  else
    _instrument = "RGS2";

  if ( tmpFileNames.size() > 1 )
    if  ( tmpFileNames[1][0].find(_instrument) == string::npos )
      _instrument = "RGS";


  _filter = tab->attribute("FILTER")->asString();
  _mjdref = tab->attribute("MJDREF")->asString();
  _object = tab->attribute("OBJECT")->asString();
  _obs_id = tab->attribute("OBS_ID")->asString();
  _exp_id = tab->attribute("EXP_ID")->asString();
  _exp_id_str = _exp_id.substr(10,12);
  _observer = tab->attribute("OBSERVER")->asString();
  _ra_obj = tab->attribute("RA_OBJ")->asReal();
  _dec_obj = tab->attribute("DEC_OBJ")->asReal();
  setSourceID(srcNo);
  dataSetServer->close(srcSet);
}

void Lightcurve::addBkgDataToFile(std::string fileName)
{
	DataSet * out = dataSetServer->open(fileName, Dal::Modify);
	Table *table = out->table("RATE");

	Column *colRate = table->addColumn("BACKV",Column::Real32,"Bkg Count Rate","count/s");
	ColumnData *dRate = colRate->data();
	real32 *r32Rate = dRate->real32Data();
	for (unsigned long i = 0; i < _counts.size(); ++i)
	    r32Rate[i] = _countRate[i];
	delete dRate;

	Column *colErrRate = table->addColumn("BACKE",Column::Real32,"BKg Count Rate Error","count/s");
	ColumnData *dErrRate = colErrRate->data();
	real32 *r32ErrRate = dErrRate->real32Data();
	for (unsigned long i = 0; i < _counts.size(); ++i)
		r32ErrRate[i] = _errCountRate[i];
	delete dErrRate;

	dataSetServer->close(out);
}


void Lightcurve::createFile(std::string fileName,bool isSrc,bool isBkgApplied)
{

  DataSet * out = dataSetServer->open(fileName, Dal::Create);
  Table *table = out->addTable("RATE", _countRate.size() );

  //Copy global attributes

  *table->addAttribute("TELESCOP","XMM Mission" )=_mission;
  *table->addAttribute( "INSTRUME",
			"RGS instrument")= _instrument;
  *table->addAttribute("FILTER","RGS has only one fixed filter" )= _filter;
  *table->addAttribute("MJDREF","1998-01-01T00:00:00 (TT) expressed in MJD (TT)" )= _mjdref;

  *table->addAttribute("OBJECT","Object name")= _object;
  *table->addAttribute("OBS_ID","Observation identifier")= _obs_id;
  *table->addAttribute("EXP_ID","Exposure identifier")= _exp_id;
  *table->addAttribute("OBSERVER","Observer")= _observer;
  *out->addAttribute("EXPIDSTR","Exp Identifier (nnn)")= _exp_id_str;
  *table->addAttribute("RA_OBJ","[deg] RA of source used in attitude correction")= _ra_obj;
  *table->addAttribute("DEC_OBJ","[deg] DEC of source used for attitude correction")= _dec_obj;
  *table->addAttribute("SRCNUM","Source number")= _srcNo;

  Column *colTime = table->addColumn("TIME",Column::Real64,"Time","s");
  ColumnData *dTime = colTime->data();
  real64 *r64Time = dTime->real64Data();
  for (unsigned long i = 0; i < _counts.size(); ++i)
    r64Time[i] = _timeGrid[i] + ((_timeGrid[i+1] - _timeGrid[i]) / 2.0);
  delete dTime;

  Column *colRate = table->addColumn("RATE",Column::Real32,"Count Rate","count/s");
  ColumnData *dRate = colRate->data();
  real32 *r32Rate = dRate->real32Data();
  for (unsigned long i = 0; i < _counts.size(); ++i)
    r32Rate[i] = _countRate[i];
  delete dRate;

  Column *colErrRate = table->addColumn("ERROR",Column::Real32,"Count Rate Error","count/s");
  ColumnData *dErrRate = colErrRate->data();
  real32 *r32ErrRate = dErrRate->real32Data();
  for (unsigned long i = 0; i < _counts.size(); ++i)
    r32ErrRate[i] = _errCountRate[i];
  delete dErrRate;



  /*
   * Source and Background counts
   */
  Column *colSrcCounts = table->addColumn("COUNTS",Column::Real32,"Counts","cts");
  ColumnData *dSrcCounts = colSrcCounts->data();
  real32 *r32SrcCounts = dSrcCounts->real32Data();
  for (unsigned long i = 0; i < _counts.size(); ++i)
    r32SrcCounts[i] = _realCounts[i];
  delete dSrcCounts;

  Column *colSrcErrCounts = table->addColumn("ERRCTS",Column::Real32,"Error in counts","cts");
  ColumnData *dSrcErrCounts = colSrcErrCounts->data();
  real32 *r32SrcErrCounts = dSrcErrCounts->real32Data();
  for (unsigned long i = 0; i < _counts.size(); ++i)
    r32SrcErrCounts[i] = _errCounts[i];
  delete dSrcErrCounts;


  /*
   * Fracexp (exposure) info
   */

  Column *colFracExp = table->addColumn("FRACEXP",Column::Real32,"FRACEXP","");
  ColumnData *dFracExp = colFracExp->data();
  real32 *r32FracExp = dFracExp->real32Data();
  for (unsigned long i = 0; i < _counts.size(); ++i)
    r32FracExp[i] = (_fracExp[i]/_numberLcAdded)/_binSize;
  delete dFracExp;

  // Write some keywords for source

  *table->addAttribute( "HDUCLASS",
			"Format conforms to OGIP/GSFC conventions")= "OGIP";
  *table->addAttribute( "HDUCLAS1",
			"File contains a spectrum")= "LIGHTCURVE";

  *table->addAttribute( "HDUVERS1",
			"Version of format")= "1.1.0";
  *table->addAttribute( "HDUCLAS3",
  			"Light curve stored as rates")= "RATE";


  if (isBkgApplied && isSrc)
  {
	  *table->addAttribute( "HDUCLAS2",
			  "File contains background subtracted rates")= "NET";
	  *table->addAttribute( "BACKAPP", "Background correction applied?") = true;
  }
  else
  {
	  *table->addAttribute( "HDUCLAS2",
			  "File contains gross rates")= "TOTAL";
	  *table->addAttribute( "BACKAPP", "Background correction applied?") = false;
  }


  *table->addAttribute( "DEADAPP", "Dead time correction applied?") = true;
  *table->addAttribute( "VIGNAPP", "Vignetting correction applied?") = false;
  *table->addAttribute( "BACKSCA", "BACKSCALE Weighted Mean") = _backScaleWeighted;

  // Rate file specific keywords.
  *table->addAttribute( "TIMEDEL",
			"Light curve bin size")= _binSize;
  *table->addAttribute( "TSTART",
			"Light curve start time")= _startTime;
  *table->addAttribute( "TSTOP",
			"Light curve end time")= _endTime;
  // Only if background substracted


  dataSetServer->close(out);
}

void Lightcurve::printToFile(string fileName)
{

  ofstream outFile;
  //std::string textFileName=stringParameter("outputfilename")+".dat";
  outFile.open(fileName.c_str());

  outFile <<fixed;
  outFile.precision(20);
  outFile.setf(ios_base::showpoint);
  for (unsigned long i = 0; i < _counts.size(); ++i)
    {
      outFile << _timeGrid[i] + ((_timeGrid[i+1] - _timeGrid[i]) / 2.0) << " "
	      << _counts[i]/_fracExp[i] << " " << _errCounts[i]/_fracExp[i] << endl;
    }
  outFile.close();

}

