// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
/*
  rgslccorr

  Authors:  Aitor Ibarra - XMM-Newton SOC (ESAC/ESA)

  Description: rgs light curve correction

  Comments:

  $Id: rgslccorr.cc,v 1.14 2013/01/16 11:44:36 aibarra Exp $
 */


/* SAS */
#include "Dal.h"
#include "DataFilter.h"
#include "param.h"
#include "PhysicalConstants.h"

/* Local */
#include "rgslccorr.h"
#include "Lccorr.h"
#include "Lightcurve.h"

#include <Tracer.h>
#include <errstr.h>
#include <vector>
#include <iomanip>

using namespace std;

string checkUnits(DataSet *);

void rgslccorr ()
{
	PROC
	//Vector to store the intermediate file names
	std::vector<std::vector<std::string> > filteredVectorNames;

	//Check that the binsize is an integer number
	real64 _binSize = realParameter("timebinsize");
	if( floor(_binSize) != _binSize ){
		errstr << "Binning factor must be an integer."<<Sas::error("BinsizeVal");}


	std::string sourceid = stringParameter("sourceid");
	// Open input event list and source list files
	vector<string> evlistSet = stringListParameter("evlist");
	// Source list
	vector<string> srclistSet = stringListParameter("srclist");

	if ( evlistSet.size() != srclistSet.size() )
		errstr << "The number of event list files and source list files are different  " <<Sas::error("WrongInputFiles");


//	if ( evlistSet.size() > 2)
//		errstr << "Only two event list are permitted. " <<Sas::error("EventListSize");

	//Check if we have to subtract background
	bool _subtractBKG =  booleanParameter("withbkgsubtraction");

	// Initial values needed for the comparison
	std::string obsid="0";
	std::string oldObsid="0";
	double tstart = 0.0;
	double tstop  = 0.0;
	//Units to check if we are dealing angstroms or beta files
	string enerUnits;
	for (unsigned int sourceFile = 0 ; sourceFile < evlistSet.size() ; sourceFile++)
	{
		std::vector<std::string> pair;
		DataSet * nameSet = dataSetServer->open(evlistSet[sourceFile], Dal::Read);
		//Check that we have two Event list and two source list and
		// also check that both belongs to the same observation
		obsid =  nameSet->attribute("OBS_ID")->asString();
		if (sourceFile != 0)
			if (obsid != oldObsid)
				errstr << "The event lists does not belong to the same observation. " <<Sas::error("WrongObservation");

		if (sourceFile == 0)
		{
			tstart = nameSet->table("EVENTS")->attribute("TSTART")->asReal();
			tstop  = nameSet->table("EVENTS")->attribute("TSTOP")->asReal();
		}

		if  (sourceFile != 0)
			if ( (tstart >= nameSet->table("EVENTS")->attribute("TSTOP")->asReal()) ||
					(tstop <= nameSet->table("EVENTS")->attribute("TSTART")->asReal() ) )
				errstr << "Overlapping time between eventlist is zero" << Sas::error("NotOverlappingTime");


		std::stringstream filteredSetName;
		filteredSetName <<  "sourceEvents_"<< nameSet->attribute("INSTRUMENT")->asString() << "_" << sourceFile+1 <<".ds";
		std::stringstream filteredBkgSetName;
		if (_subtractBKG)
			filteredBkgSetName <<"backgroundEvents_"<< nameSet->attribute("INSTRUMENT")->asString() << "_" <<sourceFile+1 <<".ds";
		//Store the name of the intermediate files that will be removed at the end of
		//the process
		pair.push_back(filteredSetName.str());
		if (_subtractBKG)
			pair.push_back(filteredBkgSetName.str());

		filteredVectorNames.push_back(pair);
		pair.clear();

		//Check the units of the EXPMAP extension to define the regions in beta or
		//Angstroms
		enerUnits = checkUnits(nameSet);

		dataSetServer->close(nameSet);
		oldObsid = obsid;



	}


	/*
	 * Final Lightcurve (needed to get the start and stop time) to get the time limits
	 */
	Lightcurve *lc = new Lightcurve(realParameter("timebinsize"));


	for (unsigned int sourceFile = 0 ; sourceFile < evlistSet.size() ; sourceFile++)
	{

		errstr<<"Cloning Source dataset ... "<<Sas::message(Msg::App, Msg::Verbose);
		DataSet *set = dataSetServer->clone(evlistSet[sourceFile], filteredVectorNames[sourceFile][0], Dal::Modify);
		errstr<<"done."<<Sas::message(Msg::App,Msg::Verbose);
		Table *srcTable = set->table("EVENTS");

		DataSet *bkgSet = 0;
		Table *bkgTable = 0;
		if (_subtractBKG)
		{
			errstr<<"Cloning Source dataset for background ... "<<Sas::message(Msg::App, Msg::Verbose);
			bkgSet = dataSetServer->clone(evlistSet[sourceFile], filteredVectorNames[sourceFile][1], Dal::Modify);
			errstr<<"done."<<Sas::message(Msg::App,Msg::Verbose);
			bkgTable = bkgSet->table("EVENTS");
		}

		/*
		 * Filter Source data and create the Ligthcurve for each ccd
		 */

		DataSet * srcSet = dataSetServer->open(srclistSet[sourceFile], Dal::Read);

		Table *table = srcSet->table(set->attribute("INSTRUMENT")->asString()+"_SRC"+sourceid+"_SPATIAL");
		Table *bkgListTable = 0;
		if (_subtractBKG)
			bkgListTable = srcSet->table(set->attribute("INSTRUMENT")->asString()+"_BACKGROUND");

		string enerCol;
		if (enerUnits == "angstrom")
			enerCol = "M_LAMBDA";
		else
			enerCol = "BETA_CORR";

		//Source expression
		std::stringstream expression;
		expression << "REGION("<<srclistSet[sourceFile]<<":"<<set->attribute("INSTRUMENT")->asString()<<"_SRC"+sourceid+"_SPATIAL,"+enerCol+",XDSP_CORR)";

		//Background expression
		std::stringstream bkgExpression;
		if (_subtractBKG)
			bkgExpression << "REGION("+srclistSet[sourceFile]+":"+set->attribute("INSTRUMENT")->asString()+"_BACKGROUND,"+enerCol+",XDSP_CORR)";


		vector<int> orders = intListParameter("orders");

		for ( unsigned int order = 1; order <= orders.size(); ++order ) {
			if (order == 1)
			{
				if (orders.size() == 2)
				{
					expression << " && ( ";
					if (_subtractBKG)
						bkgExpression << " && ( ";
				}
				else
				{
					expression << " && ";
					if (_subtractBKG)
						bkgExpression << " && ";
				}
			}
			if (order == 2)
			{
				expression << " || ";
				if (_subtractBKG)
					bkgExpression << " || ";
			}

			expression << "( REGION("+srclistSet[sourceFile]+":"<<set->attribute("INSTRUMENT")->asString()
			<<"_SRC"+sourceid+"_ORDER_" << orders[order-1] << ","+enerCol+",PI) ";

			if (_subtractBKG)
			{
				bkgExpression << "( REGION("+srclistSet[sourceFile]+":"<<set->attribute("INSTRUMENT")->asString()
				<<"_SRC"+sourceid+"_ORDER_" << orders[order-1] << ","+enerCol+",PI) ";
			}

			// Wavelength filtering
			if(booleanParameter("withfiltering"))
			{
				real64 lmin;
				real64 lmax;
				if  (stringParameter("filtering") == "wavelength")
				{
					lmin = realParameter("lambdamin");
					lmax = realParameter("lambdamax");
				}
				else
				{
					real64 emin = realParameter("energymin");
					real64 emax = realParameter("energymax");
					lmax =  PhysicalConstants::eVtimesAo / emin;
					lmin = PhysicalConstants::eVtimesAo / emax;
				}

				expression << " && (M_LAMBDA in [" << lmin*double(orders[order-1]) << ":" << lmax*double(orders[order-1]) << "]) )";
				if (_subtractBKG)
					bkgExpression << " && (M_LAMBDA in [" << lmin*double(orders[order-1]) << ":" << lmax*double(orders[order-1]) << "]) )";

			}
			else //Fixes SOC-SPR-6535
			{
				expression << " ) ";
				if (_subtractBKG)
					bkgExpression << " ) ";
			}

			if (order == 2)
			{
				expression << ")";
				if (_subtractBKG)
					bkgExpression << ")";
			}

		}
		// CCD filtering
		if (booleanParameter("withccdselection"))
		{
			std::stringstream ccdsSel;
			ccdsSel << " && (";
			vector<int> ccds = intListParameter("ccds");
			for (unsigned number = 0 ; number < ccds.size() ; ++number)
			{
				ccdsSel << "(CCDNR=="<<ccds[number];
				if (number < ccds.size()-1)
					ccdsSel << ")||";
				else
					ccdsSel << ")";
			}
			ccdsSel << ")";

			expression << ccdsSel.str();
			if (_subtractBKG)
				bkgExpression << ccdsSel.str();
		}

		errstr << "Source expression used: " << expression.str()<< Sas::message(Msg::App,Msg::Sparse) ;
		std::string expressionTotal = DataFilter::expandSelectionExpression(expression.str(), *table);

		//Source filtering
		DataFilter filter(expressionTotal);
		unsigned long selectedEvents = filter.doFilter(*srcTable);
		errstr << "selected " << selectedEvents << " rows from the source input table."
		<< Sas::message(Msg::App,Msg::Sparse);

		//Background filtering
		if (_subtractBKG)
		{
			errstr << "Background expression " << bkgExpression.str()<<Sas::message(Msg::App,Msg::Sparse)  ;

			std::string bkfFinalExpression = DataFilter::expandSelectionExpression(bkgExpression.str(), *bkgListTable);
			DataFilter bkgFilter(bkfFinalExpression);
			unsigned long bkgSelectedEvents = bkgFilter.doFilter(*bkgTable);
			errstr << "selected " << bkgSelectedEvents << " rows from the background input table."
			<< Sas::message(Msg::App,Msg::Sparse);
		}


		/*
		 * Set the limits of the final lightcurve using the
		 * all the ccds
		 */

		if (!booleanParameter("withtimeranges"))
		{
			for (unsigned int i = 1 ; i <= 9 ; i++)
			{
				std::stringstream gti;
				gti << "STDGTI";
				gti.width(2);
				gti.fill('0');
				gti << i;

				if (set->hasBlock(gti.str()))
				{
					Table *tb = set->table(gti.str());
					//lc->setTimeLimits(evCol);
					lc->setTimeLimits(tb->column("START"),tb->column("STOP"));
				}
			}
		}
		else
			lc->setTimeLimits(realParameter("timemin"),realParameter("timemax"));

		lc->createTimeGrid();

		if (_subtractBKG)
			dataSetServer->close(bkgSet);
		dataSetServer->close(srcSet);
		dataSetServer->close(set);

	}

	/*
	 * Get the time limits calculate for all the ccds
	 */
	vector<double> timeLimits = lc->getTimeLimits();
	errstr <<  std::setprecision(20);
	errstr << "Time start selected " << timeLimits[0] << Sas::message(Msg::App,Msg::Sparse);
	errstr << "Time stop selected  " << timeLimits[1] << Sas::message(Msg::App,Msg::Sparse);

	delete lc;



	/*
	 * Get the events for each CCD cloning the source dataSet an filtering
	 */

	vector<Lightcurve *> ccdLcVect;
	vector<Lightcurve *> ccdBkgLcVect;

	for (unsigned int sourceFile = 0 ; sourceFile < evlistSet.size() ; sourceFile++)
	{
		for (unsigned int iccd=1 ; iccd < 10 ; iccd++)
		{
			DataSet *ccdSet = dataSetServer->clone(filteredVectorNames[sourceFile][0], "tmpCCDSet.fits", Dal::Modify);
			Table *ccdTab = ccdSet->table("EVENTS");
			std::stringstream ccdExpression;
			ccdExpression << "CCDNR==";
			ccdExpression << iccd;
			errstr << "Selecting events from " << ccdExpression.str() << Sas::message(Msg::App,Msg::Sparse) ;
			DataFilter ccdFilter(ccdExpression.str());

			//Background
			DataSet *bkgSet = 0;
			Table *ccdBkgTable = 0;
			if (_subtractBKG)
			{
				bkgSet = dataSetServer->clone(filteredVectorNames[sourceFile][1], "tmpBkgCCDSet.fits", Dal::Modify);
				ccdBkgTable = bkgSet->table("EVENTS");
			}

			//Filter per CCD
			//Source
			unsigned long ccdSel = ccdFilter.doFilter(*ccdTab);
			errstr << "selected " << ccdSel << " rows from source file " << ccdExpression.str()
			<< Sas::message(Msg::App,Msg::Sparse);

			//Background
			if (_subtractBKG)
			{
				unsigned long bkgCcdSel = ccdFilter.doFilter(*ccdBkgTable);
				errstr << "selected " << bkgCcdSel << " rows from background file " <<  ccdExpression.str()
				<< Sas::message(Msg::App,Msg::Sparse);
			}

			//Test scanner
			DataSet * testSet = dataSetServer->open(srclistSet[sourceFile], Dal::Read);

			//if (testSet->table("SRCLIST")->hasAttribute("PRIMESRC"))
			//			srcNo = testSet->table("SRCLIST")->attribute("PRIMESRC")->asInt();
			//else
			//		errstr << "PRIMESRC keyword not found" << Sas::warning("NOPRIMESRC");

			LccorrUtils *lUtils = new LccorrUtils();
			LccorrUtils *lBKGUtils = 0;
			if (_subtractBKG)
				lBKGUtils = new LccorrUtils();

			//Exposure Maps
			for (unsigned int exp = 0 ; exp < 2 ; exp++)
			{
				std::stringstream exposureMap;
				exposureMap << "EXPMAP";
				exposureMap.width(2);
				exposureMap << exp;
				exposureMap << iccd;

				if ( ccdSet->hasBlock(exposureMap.str()) )
				{
					lUtils->exposureMap(ccdSet->array(exposureMap.str()));
					if (_subtractBKG)
						lBKGUtils->exposureMap(ccdSet->array(exposureMap.str()));
				}
			}

			errstr << "Calculating source area for " << ccdExpression.str() <<  Sas::message(Msg::App,Msg::Sparse);
			lUtils->createGrids(ccdTab, enerUnits);
			lUtils->scan(testSet->table(ccdSet->attribute("INSTRUMENT")->asString()+"_SRC"+sourceid+"_SPATIAL"));
			if (_subtractBKG)
			{
				errstr << "Calculating background area for " << ccdExpression.str() <<  Sas::message(Msg::App,Msg::Sparse);
				lBKGUtils->createGrids(ccdBkgTable,enerUnits);
				lBKGUtils->scan(testSet->table(ccdSet->attribute("INSTRUMENT")->asString()+"_BACKGROUND"));
			}


			/*
			 * Check if there is any event for this particular ccd
			 */
			if (ccdSel == 0)
			{
				errstr<<" No filtering events for ccd number " << iccd
				<<Sas::warning("NoFilteredEvents");
			}
			else
			{
				/*
				 * Column containing the ON time events
				 */
				Column *ccdCol = ccdTab->column("TIME",Dal::Read);

				// Background columns
				Column *bkgCcdCol = 0;
				if (_subtractBKG)
					bkgCcdCol = ccdBkgTable->column("TIME",Dal::Read);
				/*
				 * Create a lightcurve for each of the CCDs
				 */

				Lightcurve *ccdLc = new Lightcurve(realParameter("timebinsize"),iccd);
				Lightcurve *ccdBkgLc = 0;
				if (_subtractBKG)
					ccdBkgLc = new Lightcurve(realParameter("timebinsize"),iccd);


				if (!booleanParameter("withtimeranges"))
				{
					ccdLc->setTimeLimits(timeLimits[0],timeLimits[1]);
					if (_subtractBKG)
						ccdBkgLc->setTimeLimits(timeLimits[0],timeLimits[1]);
				}
				else
				{
					ccdLc->setTimeLimits(realParameter("timemin"),realParameter("timemax"));
					if (_subtractBKG)
						ccdBkgLc->setTimeLimits(realParameter("timemin"),realParameter("timemax"));
				}

				std::stringstream exposu;
				exposu << "EXPOSU";
				exposu.width(2);
				exposu.fill('0');
				exposu << iccd;

				std::stringstream gti;
				gti << "STDGTI";
				gti.width(2);
				gti.fill('0');
				gti << iccd;

				ccdLc->setExposuInfo(ccdSet->table(exposu.str()),ccdSet->table(gti.str()), ccdCol);
				if (_subtractBKG)
					ccdBkgLc->setExposuInfo(ccdSet->table(exposu.str()),ccdSet->table(gti.str()), bkgCcdCol);

				ccdLc->createTimeGrid();
				if (_subtractBKG)
					ccdBkgLc->createTimeGrid();

				ccdLc->setArea(lUtils->getArea());
				errstr << "Dead Time correction for "<< ccdExpression.str() <<"..." <<Sas::message(Msg::App,Msg::Sparse);
				ccdLc->createLc();
				if (_subtractBKG)
				{
					ccdBkgLc->setArea(lBKGUtils->getArea());
					ccdBkgLc->createLc();
				}
				errstr << "done. " <<  Sas::message(Msg::App,Msg::Sparse);
				ccdLc->accumCounts();
				if (_subtractBKG)
					ccdBkgLc->accumCounts();
				errstr << "Calculating  lightcurve errors... " <<  Sas::message(Msg::App,Msg::Sparse);
				ccdLc->calculateErrors();
				if (_subtractBKG)
					ccdBkgLc->calculateErrors();
				errstr << "done. " <<  Sas::message(Msg::App,Msg::Sparse);

				if (_subtractBKG)
				{
					errstr << "Backscale correction for "<< ccdExpression.str() <<"..." <<Sas::message(Msg::App,Msg::Sparse);
					ccdBkgLc->correctBackScale(ccdLc);
					errstr << "done. " <<  Sas::message(Msg::App,Msg::Sparse);
				}

				errstr << "Source: CCD "<< ccdLc->getCcdNr() << " Counts= "<< ccdLc->getTotalCounts() <<" Area= "
					<< ccdLc->getArea() << Sas::message(Msg::App,Msg::Sparse);
				if (_subtractBKG)
				{
					errstr << "Background: CCD "<< ccdBkgLc->getCcdNr() << " Counts= "<< ccdBkgLc->getTotalCounts() <<" Area= "
									<< ccdBkgLc->getArea() << Sas::message(Msg::App,Msg::Sparse);
				}

				ccdLcVect.push_back(ccdLc);
				if (_subtractBKG)
					ccdBkgLcVect.push_back(ccdBkgLc);

				delete lUtils;
				delete lBKGUtils;

			}

			if (_subtractBKG)
				dataSetServer->close(bkgSet);
			dataSetServer->close(ccdSet);

			if (_subtractBKG)
				dataSetServer->remove("tmpBkgCCDSet.fits");
			dataSetServer->remove("tmpCCDSet.fits");
		}
	}


	errstr << "Adding source lightcurves for each CCD... " <<  Sas::message(Msg::App,Msg::Sparse);
	//If only 1 CCD has been selected, create the initial _totalCCDCounts and _ccdArea vectors.
	ccdLcVect[0]->add();

	for (unsigned int i = 1 ; i < ccdLcVect.size() ; i++)
		ccdLcVect[0]->add(ccdLcVect[i]);
	errstr << "done. " <<  Sas::message(Msg::App,Msg::Sparse);

	if (_subtractBKG)
	{
		errstr << "Adding background lightcurves for each CCD... " <<  Sas::message(Msg::App,Msg::Sparse);
		//If only 1 CCD has been selected, create the initial _totalCCDCounts and _ccdArea vectors.
		ccdBkgLcVect[0]->add();

		for (unsigned int i = 1 ; i < ccdBkgLcVect.size() ; i++)
			ccdBkgLcVect[0]->add(ccdBkgLcVect[i]);
		errstr << "done. " <<  Sas::message(Msg::App,Msg::Sparse);
	}

	/*
	 * Calculate a weigthed BACKSCALE
	 */
	if (_subtractBKG)
		ccdLcVect[0]->calculateWeightedBackScale(ccdBkgLcVect[0]);


	if (_subtractBKG)
	{
		errstr << "subtracting background lightcurve... " <<  Sas::message(Msg::App,Msg::Sparse);
		ccdLcVect[0]->substract(ccdBkgLcVect[0]);
		errstr << "done. " <<  Sas::message(Msg::App,Msg::Sparse);
	}


	/*
	 * Copy the background count rates into the src lightcurve
	 */
	//ccdLcVect[0]->copyBkgInfo(ccdBkgLcVect[0]);

	//Convert string to int
	errstr << "Creating output file... " <<  Sas::message(Msg::App,Msg::Sparse);
	ccdLcVect[0]->setDefaultKeywords(filteredVectorNames, std::atoi( sourceid.c_str() ));
	ccdLcVect[0]->createFile(stringParameter("outputsrcfilename"),true,_subtractBKG);
	if (_subtractBKG)
	{
	  ccdBkgLcVect[0]->setDefaultKeywords(filteredVectorNames,std::atoi(sourceid.c_str()));
		ccdBkgLcVect[0]->createFile(stringParameter("outputbkgfilename"),false,_subtractBKG);
		ccdBkgLcVect[0]->addBkgDataToFile(stringParameter("outputsrcfilename"));

	}
	errstr << "done. " <<  Sas::message(Msg::App,Msg::Sparse);

	//ccdLcVect[0]->printToFile("source_bkg.dat");

	errstr << "Deleting temporary files and cleaning memory... " <<  Sas::message(Msg::App,Msg::Sparse);
	//Remove the temporary files
	for (unsigned int i = 0 ; i < filteredVectorNames.size(); i++)
		for (unsigned int j = 0 ; j < filteredVectorNames[i].size(); j++)
			dataSetServer->remove(filteredVectorNames[i][j]);

	for (unsigned int i = 0 ; i < ccdLcVect.size() ; i++)
		delete ccdLcVect[i];

	if (_subtractBKG)
	{
		for (unsigned int i = 0 ; i < ccdBkgLcVect.size() ; i++)
			delete ccdBkgLcVect[i];
	}
	errstr << "done. " <<  Sas::message(Msg::App,Msg::Sparse);

	ENDPROC
}


string  checkUnits(DataSet *nameSet)
{
	string unit="";

	//Exposure Maps
	for (unsigned int iccd=1 ; iccd < 10 ; iccd++)
	{
		for (unsigned int exp = 0 ; exp < 2 ; exp++)
		{
			std::stringstream exposureMap;
			exposureMap << "EXPMAP";
			exposureMap.width(2);
			exposureMap << exp;
			exposureMap << iccd;

			if ( nameSet->hasBlock(exposureMap.str()) )
			{
				unit = nameSet->array(exposureMap.str())->attribute("CUNIT1")->asString();
				return unit;
			}
		}
	}

	if (unit == "")
	{
		errstr << "No EXPMAP extension found " << Sas::error("NoEXPMAPFound");
	}
	return unit;
}

